/*
 * modified for lecture by Fabian Fuchs (KIT, Karlsruhe) based on:
 *  
 Copyright (c) 2007, Distributed Computing Group (DCG)
                    ETH Zurich
                    Switzerland
                    dcg.ethz.ch

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

 - Neither the name 'Sinalgo' nor the names of its contributors may be
   used to endorse or promote products derived from this software
   without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package projects.uebung3solution;


import javax.swing.JOptionPane;

import sinalgo.configuration.Configuration;
import sinalgo.configuration.CorruptConfigurationEntryException;
import sinalgo.gui.helper.NodeSelectionHandler;
import sinalgo.nodes.Node;
import sinalgo.runtime.AbstractCustomGlobal;
import sinalgo.runtime.Global;
import sinalgo.runtime.Main;
import sinalgo.tools.Tools;
import sinalgo.tools.logging.Logging;

/**
 * This class holds customized global state and methods for the framework. 
 * The only mandatory method to overwrite is 
 * <code>hasTerminated</code>
 * <br>
 * Optional methods to override are
 * <ul>
 * <li><code>customPaint</code></li>
 * <li><code>handleEmptyEventQueue</code></li>
 * <li><code>onExit</code></li>
 * <li><code>preRun</code></li>
 * <li><code>preRound</code></li>
 * <li><code>postRound</code></li>
 * <li><code>checkProjectRequirements</code></li>
 * </ul>
 * @see sinalgo.runtime.AbstractCustomGlobal for more details.
 * <br>
 * In addition, this class also provides the possibility to extend the framework with
 * custom methods that can be called either through the menu or via a button that is
 * added to the GUI. 
 */
public class CustomGlobal extends AbstractCustomGlobal{
	
	Logging log = Logging.getLogger("u1_log.txt");
	
	// The user can optionally specify exitAfter in the config file to indicate after how many rounds the simulation should stop. 
	boolean exitAfterFixedRounds = false;
	int exitAfterNumRounds;
	{
		if(Configuration.hasParameter("exitAfter")) {
			try {
				exitAfterFixedRounds = Configuration.getBooleanParameter("exitAfter");
			} catch (CorruptConfigurationEntryException e1) {
				Tools.fatalError("The 'exitAfter' needs to be a valid boolean.");
			}
			if(exitAfterFixedRounds) {
				try {
					exitAfterNumRounds = Configuration.getIntegerParameter("exitAfter/rounds");
				} catch (CorruptConfigurationEntryException e) {
					Tools.fatalError("The 'exitAfter/rounds' parameter specifies the maximum time the simulation runs. It needs to be a valid integer.");
				}
			}
		} else {
			exitAfterFixedRounds = false;
		}
	}
	
	
	/* (non-Javadoc)
	 * @see runtime.AbstractCustomGlobal#hasTerminated()
	 */
	public boolean hasTerminated(){
		if(exitAfterFixedRounds) {
			return exitAfterNumRounds <= Global.currentTime;
		}

		if(Tools.isSimulationInGuiMode()) {
			return false; // in GUI mode, have the user decide when to stop.
		} else {
			return Global.currentTime > 100000; // stop after x rounds 
		}
	}

	
	/* (non-Javadoc)
	 * @see runtime.AbstractCustomGlobal#preRun()
	 */
	public void preRun() {
				
	}
	
	public void preRound() {
		//do something before each round?
	}
	
	public void postRound() {
		double dt = System.currentTimeMillis() - Global.startTimeOfRound.getTime();
		log.logln("Round " + (int)(Global.currentTime) + " time: "  + dt + " example text");
		this.preRound();
	}
	
	
	/**
	 * Transmit message from a node to another node
	 */
	@AbstractCustomGlobal.GlobalMethod(menuText="Start Routing")
	public void startRoutingMenu() {

		//quite a hack, beware!
		/* refine nodeSelectionHandler that stores the node and (optionally) a second handler */
		class NodeSelectionHandlerAndStore implements NodeSelectionHandler {
			private Node node;
			private boolean wasHandled;
			
			/* stores another handler. The handlers information must be available once
			 * this handler got its information. As this handler gets the information
			 * only once the user clicked (and the simulator must move on for the user
			 * interface to be clickable), we put one handler inside the other so that 
			 * second handler can access the first handlers information and ultimately 
			 * measure the distance.
			 */
			private NodeSelectionHandlerAndStore previous; 
			
			//constructors
			NodeSelectionHandlerAndStore() {
				wasHandled = false;
			}
			NodeSelectionHandlerAndStore(NodeSelectionHandlerAndStore prev) {
				wasHandled = false;
				previous = prev;
			}
			public boolean wasHandled() {
				return wasHandled;
			}
			public Node getNode() {
				return node;
			}
			public void handleNodeSelectedEvent(Node n) { //n = null implies abort by user
				/* if this is the first handler, skip this and store wasHandled and true
				 * otherwise (and the first handle is set and was handled), get both nodes
				 * and compute the distance. 
				 */
				if (previous != null && previous.wasHandled()) {
					Node n1 = n;
					Node n2 = previous.getNode();
					if (n1 == null || n2 == null) {
						JOptionPane.showMessageDialog(null, "'selection aborted by user'", "Abort" , JOptionPane.INFORMATION_MESSAGE);
					}
					JOptionPane.showMessageDialog(null, "'Distance = " + n1.getPosition().distanceTo(n2.getPosition()) + "'", "Distance between Node "+n1.ID+" and "+n2.ID , JOptionPane.INFORMATION_MESSAGE);
				}
				wasHandled = true;
				node = n;
			}
		}
		/* end of class definition */

		//create node selectors
		NodeSelectionHandlerAndStore nsh1 = new NodeSelectionHandlerAndStore();
		NodeSelectionHandlerAndStore nsh2 = new NodeSelectionHandlerAndStore(nsh1);
		
		//let the user select the nodes
		/* data structure to organize these events is a stack, thus second node first */
		Tools.getNodeSelectedByUser(nsh2, "Select the second node.");
		Tools.getNodeSelectedByUser(nsh1, "Select the first node.");
	}
	
	
	public void checkProjectRequirements(){
		//does the configuration fit the projects requirements?
		//Otherwise, call Main.fatalError("error message")
	}
	
	public void onExit() {
		// perform some cleanup operations here
	}
	
}
